#ifndef GBRL_H
#define GBRL_H

#include <string>
#include <tuple>

#include "node.h"
#include "optimizer.h"
#include "loss.h"
#include "split_candidate_generator.h"
#include "types.h"

#ifdef USE_CUDA
#include "cuda_types.h"
#endif


class GBRL {
    public:
        GBRL(int output_dim, int policy_dim, int max_depth, int min_data_in_leaf, 
             int n_bins, int par_th, float cv_beta, scoreFunc split_score_func, generatorType generator_type, bool use_control_variates, 
             int batch_size, growPolicy grow_policy, int verbose, deviceType device);
        GBRL(int output_dim, int policy_dim, int max_depth, int min_data_in_leaf, 
             int n_bins, int par_th, float cv_beta, std::string split_score_func, std::string generator_type, bool use_control_variates, 
             int batch_size, std::string grow_policy, int verbose, std::string device);
        GBRL(const std::string& filename);
        GBRL(GBRL& other);
        ~GBRL();
        static bool cuda_available();
        void to_device(deviceType device);
        std::string get_device();
        int saveToFile(const std::string& filename);
        int loadFromFile(const std::string& filename);

        void fit(const float *obs, const char *categorical_obs, float *grads, const int n_samples, const int n_num_features, const int n_cat_features);
#ifdef USE_CUDA
        void _fit_gpu(dataSet *dataset);
        float _fit_sl_gpu(dataSet *datasett, float *targets, const int n_iterations);
#endif
        float fit_sl(float *obs, char *categorical_obs, float *targets, int iterations, const int n_samples, const int n_num_features, const int n_cat_features, bool shuffle = true, std::string _loss_type = "MultiRMSE");
        void set_bias(float *bias, const int output_dim);
        float* get_bias();
        float* predict(const float *obs, const char *categorical_obs, const int n_samples, const int n_num_features, const int n_cat_features, int start_tree_idx, int stop_tree_idx);
        void predict(const float *obs, const char *categorical_obs, float *start_preds, const int n_samples, const int n_num_features, const int n_cat_features, int start_tree_idx = 0, int stop_tree_idx = 0);
        
        std::tuple<float, float> get_scheduler_lrs();

        int get_num_trees();
        int get_iteration();

        void set_optimizer(optimizerAlgo algo, schedulerFunc scheduler_func, float init_lr, float stop_lr, int T, 
                           float beta_1, float beta_2, float eps, float shrinkage);

        void print_tree(int tree_idx);
        #ifdef USE_GRAPHVIZ
        void plot_tree(int tree_idx, const std::string &filename);
        #endif

        ensembleData *edata;
        ensembleMetaData *metadata;
        std::vector<Optimizer*> opts;
        deviceType device = unspecified;
        bool parallel_predict = true;
#ifdef USE_CUDA
        SGDOptimizerGPU** cuda_opt = nullptr;
        int n_cuda_opts;
#endif 
};





#endif 